{ TIEButton component version 2.0, copyright  A. Meeder, 1997

  The idea for TIEButton came from Microsoft's Internet Explorer 3.0 (this is where 'IE' stands for).
  The buttons on the button/task-bar behave different from normal 'standard' Windows-buttons.

  This version is a totaly rewritten version of my former TIEButton-component. When you used the
  older version, you will have to delete that one from the component-palet and also from your programs.

  This control now has many new features described below:

  properties
    AllowAllUp (boolean, default false) same as with 'TSpeedButton' from Delphi
    BorderWidth (default 1, min. 1 / max. 5)
    ColorHighLight (default clBtnHighLight, see also ColorShadow)
    ColorShadow (default clBtnShadow, see also ColorHighLight)
    Down (default false, button-state)
    Glyph (default Empty, see also NumGlyphs)
    GroupIndex (default 0, button behaves like IE 3.0 button, > 0 button behaves like MS Office 97 speedbuttons)
    InactiveStyle (default dsFocus, style of inavtive glyph when no one is available and GroupIndex = 0)
    LayOut (default blGlyphTop, also Bottom, Left and Right)
    MenuMark (default false, only applicable when FGroupIndex = 0, see also PopupMenu)
    NumGlyphs (default 2, or Glyph.Width div Glyph.Height min. 1 max. 3)
    PopupMenu (default nil, You will only have to select menu and it will work correct...)
    ShowCaption (default true)
    Spacing (default 4, margin between Caption and Glyph)
    TransparentColor (default clOlive, when Glyph <> nil default lower left pixelcolor, can be selected)

  events
    OnMouseEnter
    OnMouseLeave

  If you have suggestions about TIEButton or make changes, please mail me: ameeder@dds.nl
}

unit IEBtn;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, ExtCtrls, Menus;

type
  TButtonLayout = (blGlyphLeft, blGlyphRight, blGlyphTop, blGlyphBottom);
  TBorderWidth = 1..5;
  TSpacing = 0..MAXINT;
  TNumGlyphs = 1..3;

  TMouseEnter = procedure (Sender: TControl; var Message: TMessage) of object;
  TMouseLeave = procedure (Sender: TControl; var Message: TMessage) of object;

  TIEButton = class(TGraphicControl)
  private
    FAllowAllUp: boolean;
    FLayOut: TButtonLayOut;
    FColorHighLight: TColor;
    FColorShadow: TColor;
    FBorderWidth: TBorderWidth;
    FDown: boolean;
    FGroupIndex: integer;
    FGlyph: TBitmap;
    FCurrentGlyph: TBitmap;
    FImageList: TImageList;
    FInactiveStyle: TDrawingStyle;
    FSpacing: TSpacing;
    FMenuMark: boolean;
    FNumGlyphs: TNumGlyphs;
    FPopupMenu: TPopupMenu;
    FShowCaption: boolean;
    FTransparentColor: TColor;
    FVisible: boolean;

    FMouseEnter: TMouseEnter;
    FMouseLeave: TMouseLeave;

    FMark: TBitmap;
    FHasMouse: boolean;
    FMouseDown: boolean;
    FMouseIsDown: boolean;

    function GetDown: boolean;
    procedure SetDown(value: boolean);
    procedure SetAllowAllUp(value: boolean);
    procedure SetLayOut(value: TButtonLayOut);
    procedure SetBorderWidth(value: TBorderWidth);
    procedure SetColor(index: integer; value: TColor);
    procedure SetGlyph(value: TBitmap);
    procedure SetGroupIndex(value: integer);
    procedure SetInactiveStyle(value: TDrawingStyle);
    procedure SetSpacing(value: TSpacing);
    procedure SetMenuMark(value: boolean);
    procedure SetNumGlyphs(value: TNumGlyphs);
    procedure SetPopupMenu(value: TPopupMenu);
    procedure SetShowCaption(value: boolean);
    procedure SetTransparentColor(value: TColor);
    procedure SetVisible(value: boolean);

    procedure SetCurrentGlyph;
    procedure CurrentLayOut(var txtRect, bitRect: TRect);
    procedure UpdateExclusive;
    procedure CreateBrushPattern;
    procedure MakeTransparent(DrwStyle: TDrawingStyle; var LeftPt, TopPt: integer);
  protected
    procedure CMButtonPressed(var Message: TMessage); message CM_BUTTONPRESSED;
    procedure CMColorChanged(var Message: TMessage); message CM_COLORCHANGED;
    procedure CMFontChanged(var Message: TMessage); message CM_FONTCHANGED;
    procedure CMEnabledChanged(var Message: TMessage); message CM_ENABLEDCHANGED;
    procedure CMTextChanged(var Message: TMessage); message CM_TEXTCHANGED;
    procedure CMMouseEnter(var Message: TMessage); message CM_MOUSEENTER;
    procedure CMMouseLeave(var Message: TMessage); message CM_MOUSELEAVE;
    procedure CMSysColorChange(var Message: TMessage); message CM_SYSCOLORCHANGE;
    procedure WMMouseMove(var Message: TWMMouseMove); message WM_MOUSEMOVE;
    procedure WMLButtonDblClk(var Message: TWMLButtonDown); message WM_LBUTTONDBLCLK;
    procedure Loaded; override;
    procedure Notification(AComponent: TComponent; Operation: TOperation); override;
    procedure Paint; override;
    procedure MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
    procedure MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer); override;
  public
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;
  published
    property AllowAllUp: boolean read FAllowAllUp write SetAllowAllUp default False;
    property BorderWidth: TBorderWidth read FBorderWidth write SetBorderWidth default 1;
    property ColorHighLight: TColor index 1 read FColorHighLight write SetColor default clBtnHighLight;
    property ColorShadow: TColor index 2 read FColorShadow write SetColor default clBtnShadow;
    property Caption;
    property Color;
    property Down: boolean read GetDown write SetDown default False;
    property Enabled;
    property Font;
    property Glyph: TBitmap read FGlyph write SetGlyph;
    property GroupIndex: integer read FGroupIndex write SetGroupIndex default 0;
    property InactiveStyle: TDrawingStyle read FInactiveStyle write SetInactiveStyle default dsFocus;
    property LayOut: TButtonLayOut read FLayOut write SetLayOut default blGlyphTop;
    property MenuMark: boolean read FMenuMark write SetMenuMark default False;
    property NumGlyphs: TNumGlyphs read FNumGlyphs write SetNumGlyphs default 2;
    property ParentColor;
    property ParentFont;
    property PopupMenu: TPopupMenu read FPopupMenu write SetPopupMenu;
    property ShowCaption: boolean read FShowCaption write SetShowCaption default True;
    property ShowHint;
    property Spacing: TSpacing read FSpacing write SetSpacing default 4;
    property TransparentColor: TColor read FTransparentColor write SetTransparentColor default clOlive;
    property Visible: boolean read FVisible write SetVisible default True;

    property OnClick;
    property OnDblClick;
    property OnDragDrop;
    property OnDragOver;
    property OnEndDrag;
    property OnMouseDown;
    property OnMouseEnter: TMouseEnter read FMouseEnter write FMouseEnter;
    property OnMouseLeave: TMouseLeave read FMouseLeave write FMouseLeave;
    property OnMouseMove;
    property OnMouseUp;
    property OnStartDrag;
  end;

procedure Register;

implementation

{$R IEBTN.RES}

var
  Pattern: TBitmap = nil;

procedure TIEButton.MakeTransparent(DrwStyle: TDrawingStyle; var LeftPt, TopPt: integer);
var
  ImageIndex: integer;
begin
  FImageList := TImageList.Create(self);
  try
    FImageList.CreateSize(FGlyph.Width div FNumGlyphs, FGlyph.Height);
    ImageIndex := FImageList.AddMasked(FCurrentGlyph, FTransparentColor);
    FImageList.DrawingStyle := DrwStyle;
    FImageList.Draw(Canvas, LeftPt, TopPt, ImageIndex);
  finally
    FImageList.Free;
  end;
end;

procedure TIEButton.CreateBrushPattern;
var
  X, Y: Integer;
begin
  Pattern := TBitmap.Create;
  Pattern.Width := 8;
  Pattern.Height := 8;
  with Pattern.Canvas do
  begin
    Brush.Style := bsSolid;
    Brush.Color := Color;
    FillRect(Rect(0, 0, Pattern.Width, Pattern.Height));
    for Y := 0 to 7 do
      for X := 0 to 7 do
        if (Y mod 2) = (X mod 2) then
          Pixels[X, Y] := clWhite;
  end;
end;

procedure TIEButton.CurrentLayOut(var txtRect, bitRect: TRect);
var
  hBit, vBit, hTxt, vTxt, HalfSpacing: integer;
begin
  hBit := bitRect.Right - bitRect.Left;
  vBit := bitRect.Bottom - bitRect.Top;
  hTxt := txtRect.Right - txtRect.Left;
  vTxt := txtRect.Bottom - txtRect.Top;

  case FLayOut of
    blGlyphBottom, blGlyphTop: begin
                       bitRect.Left := (Width - hBit) div 2;
                       txtRect.Left := (Width - hTxt) div 2;
                       bitRect.Top := (Height - (vBit + vTxt + FSpacing)) div 2;
                       txtRect.Top := bitRect.Top + vBit + FSpacing;
                     end;
    blGlyphLeft, blGlyphRight: begin
                       bitRect.Top := (Height - vBit) div 2;
                       txtRect.Top := (Height - vTxt) div 2;
                       bitRect.Left := (Width - (hBit + hTxt + FSpacing)) div 2;
                       txtRect.Left := bitRect.Left + hBit + FSpacing;
                     end;
  end;

  bitRect.Right := bitRect.Left + hBit;
  bitRect.Bottom := bitRect.Top + vBit;
  txtRect.Right := txtRect.Left + hTxt;
  txtRect.Bottom := txtRect.Top + vTxt;

  if FMouseDown and FHasMouse then
  begin
    OffsetRect(bitRect, 1, 1);
    OffsetRect(txtRect, 1, 1);
  end;
end;

procedure TIEButton.SetCurrentGlyph;
var
  SourceRect, DestRect: TRect;
begin
  if not FGlyph.Empty then
  begin
    Case FNumGlyphs of
      1: SourceRect := Rect(0, 0, FGlyph.Width, FGlyph.Height);
      2: if Enabled then
           SourceRect := Rect(0, 0, FGlyph.Width div FNumGlyphs, FGlyph.Height)
         else
           SourceRect := Rect(FGlyph.Width div FNumGlyphs, 0, FGlyph.Width, FGlyph.Height);
      3: if Enabled and not FHasMouse then
           SourceRect := Rect((FGlyph.Width div FNumGlyphs) * 2, 0, FGlyph.Width, FGlyph.Height)
         else if Enabled and FHasMouse then
           SourceRect := Rect(0, 0, FGlyph.Width div FNumGlyphs, FGlyph.Height)
         else
           SourceRect := Rect(FGlyph.Width div FNumGlyphs, 0, (FGlyph.Width div FNumGlyphs) * 2, FGlyph.Height);
    end;

    DestRect := Rect(0, 0, FGlyph.Width div FNumGlyphs, FGlyph.Height);
    FCurrentGlyph.Width := FGlyph.Width div FNumGlyphs;
    FCurrentGlyph.Height := FGlyph.Height;
    FCurrentGlyph.Canvas.CopyRect(DestRect, FGlyph.Canvas, SourceRect);
  end;
end;

procedure TIEButton.Paint;
var
  bitRect, txtRect, glyphRect: TRect;
  BtnRect, MarkRect, BtnFace: TRect;
begin
  inherited Paint;

  BtnRect := ClientRect;
  BtnFace := BtnRect;
  InflateRect(BtnFace, -FBorderWidth, - FBorderWidth);
  with canvas do
  begin
    Brush.style := bsSolid;
    Brush.color := Color;

    if (FGroupIndex <> 0) and FMouseDown then
    begin
      if Pattern = nil then CreateBrushPattern;
      Brush.Bitmap := Pattern;
      FillRect(BtnFace);
    end
    else
      FillRect(BtnFace);

    if not (FVisible) and not (csDesigning in ComponentState) then Exit;

    Font.Assign(Self.Font);

    SetCurrentGlyph;
    txtRect := Rect(0, 0, TextWidth(Caption), TextHeight(Caption));
    bitRect := Rect(0, 0, FCurrentGlyph.Width, FCurrentGlyph.Height);
    glyphRect := bitRect;

    if FLayOut in [blGlyphTop, blGlyphLeft] then
      CurrentLayOut(txtRect, bitRect)
    else
      CurrentLayOut(bitRect, txtRect);

    Brush.Style := bsClear;
    if Enabled and ShowCaption then
      TextRect(txtRect, txtRect.Left, txtRect.Top, Caption)
    else if ShowCaption then
    begin
      OffsetRect(txtRect, 1, 1);
      Font.Color := clBtnHighlight;
      TextRect(txtRect, txtRect.Left, txtRect.Top, Caption);
      OffsetRect(txtRect, -1, -1);
      Font.Color := clBtnShadow;
      TextRect(txtRect, txtRect.Left, txtRect.Top, Caption);
    end;

    if not FGlyph.Empty then
    begin
      if (FGroupIndex > 0) and (FMouseDown) then
        MakeTransparent(dsTransparent, bitRect.Left, bitRect.Top)
      else if (FNumGlyphs < 3) and not (FHasMouse) and (Enabled) and (FGroupIndex = 0) then
        MakeTransparent(FInactiveStyle, bitRect.Left, bitRect.Top)
      else
      begin
        Brush.Style := bsSolid;
        Brush.Color := Color;
        BrushCopy(bitRect, FCurrentGlyph, glyphRect, FTransparentColor);
      end;
    end;

    if FMenuMark and (FLayOut in [blGlyphTop, blGlyphBottom]) then
    begin
      MarkRect := Rect(bitRect.Right+(FSpacing div 2), bitRect.Top+(FCurrentGlyph.Height div 2),
                       bitRect.Right+(FSpacing div 2)+FMark.Width, bitRect.Top+(FCurrentGlyph.Height div 2)+FMark.Height);
      BrushCopy(MarkRect, FMark, Rect(0, 0, FMark.Width, FMark.Height), clSilver);
    end;

    Brush.Style := bsSolid;
    Pen.Style := psSolid;
    if (csDesigning in ComponentState) or (FHasMouse) then
    begin
      if FMouseDown then
        Frame3D(canvas, BtnRect, FColorShadow, FColorHighLight, FBorderWidth)
      else
        Frame3D(canvas, BtnRect, FColorHighLight, FColorShadow, FBorderWidth);
    end
    else
    begin
      Frame3D(canvas, BtnRect, Color, Color, FBorderWidth);
    end;
  end;
end;

procedure TIEButton.UpdateExclusive;
var
  Msg: TMessage;
begin
  if (FGroupIndex <> 0) and (Parent <> nil) then
  begin
    Msg.Msg := CM_BUTTONPRESSED;
    Msg.WParam := FGroupIndex;
    Msg.LParam := Longint(Self);
    Msg.Result := 0;
    Parent.Broadcast(Msg);
  end;
end;

procedure TIEButton.CMButtonPressed(var Message: TMessage);
var
  Sender: TIEButton;
begin
  if Message.WParam = FGroupIndex then
  begin
    Sender := TIEButton(Message.LParam);
    if Sender <> Self then
    begin
      FMouseDown := false;
      FHasMouse := false;
      Repaint;
    end;
    FAllowAllUp := Sender.AllowAllUp;
  end;
end;

procedure TIEButton.CMColorChanged(var Message: TMessage);
begin
  Repaint;
end;

procedure TIEButton.CMFontChanged(var Message: TMessage);
begin
  Repaint;
end;

procedure TIEButton.CMEnabledChanged(var Message: TMessage);
begin
  if FMenuMark then
  begin
    if Enabled then
      FMark.Handle := LoadBitmap(HInstance, 'EPMENU')
    else
      FMark.Handle := LoadBitmap(HInstance, 'DPMENU');
  end;
  Repaint;
end;

procedure TIEButton.CMMouseEnter(var Message: TMessage);
begin
  if Enabled and FVisible then
  begin
    FHasMouse := true;
    Repaint;
    if Assigned(FMouseEnter) then FMouseEnter(Self, Message);
  end;
end;

procedure TIEButton.CMMouseLeave(var Message: TMessage);
begin
  if Enabled and FVisible then
  begin
    if (FGroupIndex <> 0) and FMouseDown then
      FHasMouse := true
    else
      FHasMouse := false;

    Repaint;
    if Assigned(FMouseLeave) then FMouseLeave(Self, Message);
  end;
end;

procedure TIEButton.CMSysColorChange(var Message: TMessage);
begin
  Repaint;
end;

procedure TIEButton.CMTextChanged(var Message: TMessage);
begin
  if Caption = '' then FSpacing := 0;
  Repaint;
end;

procedure TIEButton.WMMouseMove(var Message: TWMMouseMove);
var
  CurrentPoint: TPoint;
begin
  CurrentPoint.X := Message.XPos;
  CurrentPoint.Y := Message.YPos;
  if not FHasMouse and PtInRect(GetClientRect, CurrentPoint) then
  begin
    FHasMouse := true;
    Repaint;
  end;
  inherited;
end;

procedure TIEButton.WMLButtonDblClk(var Message: TWMLButtonDown);
begin
  inherited;
  FMouseDown := false;
end;

procedure TIEButton.MouseDown(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
var
  PopupPoint: TPoint;
begin
  if (Button = mbLeft) and Enabled and FVisible then
  begin
    FMouseIsDown := FMouseDown;
    if Assigned(FPopupMenu) then
    begin
      FMouseDown := true;
      FHasMouse := true;
      Repaint;
      PopupPoint.X := -1;
      PopupPoint.Y := Height;
      PopupPoint := ClientToScreen(PopupPoint);
      FPopupMenu.Popup(PopupPoint.X, PopupPoint.Y);
      FHasMouse := false;
      FMouseDown := false;
      Repaint;
    end
    else
    begin
      FMouseDown := true;
      Repaint;
    end;
    UpdateExclusive;
    inherited MouseDown(Button, Shift, X, Y);
  end;
end;

procedure TIEButton.MouseUp(Button: TMouseButton; Shift: TShiftState; X, Y: Integer);
begin
  if (Button = mbLeft) and Enabled and FVisible then
  begin
    if (FGroupIndex <> 0) then
    begin
      if FAllowAllUp and FMouseIsDown then
        FMouseDown := False
      else
        FMouseDown := True;
    end
    else
      FMouseDown := False;

    Repaint;
    inherited MouseUp(Button, Shift, X, Y);
  end;
end;

function TIEButton.GetDown: boolean;
begin
  Result := FMouseDown;
end;

procedure TIEButton.SetDown(value: boolean);
begin
  if FGroupIndex = 0 then Value := False;
  if Value <> FDown then
  begin
    if FDown and (not FAllowAllUp) then Exit;
    FDown := Value;
    FHasMouse := Value;
    FMouseDown := Value;
    if Value then UpdateExclusive;
    Repaint;    
  end;
end;

procedure TIEbutton.SetGroupIndex(value: integer);
begin
  if value <> FGroupIndex then
  begin
    FGroupIndex := value;
    if value > 0 then
    begin
      FMenuMark := false;
      FPopupMenu := nil;
    end;
  end;
end;

procedure TIEButton.SetLayOut(value: TButtonLayOut);
begin
  if value <> FLayOut then
  begin
    FLayOut := value;
    Repaint;
  end;
end;

procedure TIEButton.SetAllowAllUp(Value: Boolean);
begin
  if FAllowAllUp <> Value then
  begin
    FAllowAllUp := Value;
    UpdateExclusive;
  end;
end;

procedure TIEButton.SetBorderWidth(value: TBorderWidth);
begin
  if value <> FBorderWidth then
  begin
    FBorderWidth := value;
    Repaint;
  end;
end;

procedure TIEButton.SetColor(index: integer; value: TColor);
begin
  case index of
    1: if value <> FColorHighLight then
       begin
         FColorHighLight := value;
         Repaint;
       end;
    2: if value <> FColorShadow then
       begin
         FColorShadow := value;
         Repaint;
       end;
  end;
end;

procedure TIEButton.SetGlyph(value: TBitmap);
begin
  if value <> FGlyph then
  begin
    FGlyph.Assign(value);
    if not FGlyph.Empty then
    begin
      if FGlyph.Width mod FGlyph.Height = 0 then
      begin
        FNumGlyphs := FGlyph.Width div FGlyph.Height;
        if FNumGlyphs > 3 then FNumGlyphs := 1;
      end;
      FTransparentColor := FGlyph.Canvas.Pixels[0, FGlyph.Height - 1];
    end;
    Repaint;
  end;
end;

procedure TIEButton.SetInactiveStyle(value: TDrawingStyle);
begin
  if value <> FInactiveStyle then
  begin
    FInactiveStyle := value;
    Repaint;
  end;
end;

procedure TIEButton.SetSpacing(value: TSpacing);
begin
  if value <> FSpacing then
  begin
    FSpacing := value;
    Repaint;
  end;
end;

procedure TIEButton.SetMenuMark(value: boolean);
begin
  if (value <> FMenuMark) then
    FMenuMark := value;

  if FMenuMark then
    if Enabled then
      FMark.Handle := LoadBitmap(HInstance, 'EPMENU')
    else
      FMark.Handle := LoadBitmap(HInstance, 'DPMENU');

  if FMenuMark then FGroupIndex := 0;

  Repaint;
end;

procedure TIEButton.SetNumGlyphs(value: TNumGlyphs);
begin
  if value <> FNumGlyphs then
  begin
    FNumGlyphs := value;
    Repaint;
  end;
end;

procedure TIEButton.SetPopupMenu(value: TPopupMenu);
begin
  FPopupMenu := Value;
  if Value <> nil then
  begin
    Value.FreeNotification(Self);
    FGroupIndex := 0;
  end;
end;

procedure TIEButton.SetShowCaption(value: boolean);
begin
  if value <> FShowCaption then
  begin
    FShowCaption := value;
    Repaint;
  end;
end;

procedure TIEbutton.SetTransparentColor(value: TColor);
begin
  if value <> FTransparentColor then
  begin
    FTransparentColor := value;
    Repaint;
  end;
end;

procedure TIEButton.SetVisible(value: boolean);
begin
  if value <> FVisible then
  begin
    FVisible := value;
    FHasMouse := false;
    FMouseDown := false;
    if not (csDesigning in ComponentState) then Repaint;
  end;
end;

procedure TIEButton.Notification(AComponent: TComponent; Operation: TOperation);
begin
  inherited Notification(AComponent, Operation);
  if (Operation = opRemove) and (AComponent = FPopupMenu) then FPopupMenu := nil;
end;

procedure TIEButton.Loaded;
begin
  inherited Loaded;
  if FMenuMark then
  begin
    if Enabled then
      FMark.Handle := LoadBitmap(HInstance, 'EPMENU')
    else
      FMark.Handle := LoadBitmap(HInstance, 'DPMENU');
  end;
  if not (csDesigning in ComponentState) and Assigned(FPopupMenu) then
    FPopupMenu.AutoPopup := false;

  if Down then
  begin
    FHasMouse := true;
    FMouseDown := true;
  end;

  Repaint;
end;

constructor TIEButton.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  SetBounds(0, 0, 55, 45);
  ControlStyle := ControlStyle + [csSetCaption, csCaptureMouse, csOpaque, csDoubleClicks];
  FColorHighLight := clBtnHighLight;
  FColorShadow := clBtnShadow;
  FHasMouse := false;
  FMouseDown := false;
  FGlyph := TBitmap.Create;
  FCurrentGlyph := TBitmap.Create;
  FMark := TBitmap.Create;
  FLayOut := blGlyphTop;
  FAllowAllUp := False;
  FBorderWidth := 1;
  FDown := False;
  FGroupIndex := 0;
  FInactiveStyle := dsFocus;
  FSpacing := 4;
  FMenuMark := False;
  FNumGlyphs := 2;
  FShowCaption := True;
  FTransparentColor := clOlive;
  FVisible := True;
  ParentFont := True;
end;

destructor TIEButton.Destroy;
begin
  FGlyph.Free;
  FCurrentGlyph.Free;
  FMark.Free;
  inherited Destroy;
end;

procedure Register;
begin
  RegisterComponents('Win95', [TIEButton]);
end;

end.
